@@ -82,12 +82,15 @@ struct allocation_node {
 };
 
 struct allocation {
-	uint64_t stack_id;
+	int64_t stack_id;
 	size_t size;
 	size_t count;
 	struct allocation_node* allocations;
 };
 
+#define OPT_PERF_MAX_STACK_DEPTH	1 /* --perf-max-stack-depth */
+#define OPT_STACK_STORAGE_SIZE		2 /* --stack-storage-size */
+
 #define __ATTACH_UPROBE(skel, sym_name, prog_name, is_retprobe) \
 	do { \
 		char sym[32]; \
@@ -123,6 +126,17 @@ struct allocation {
 #define ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name) __ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name, false)
 #define ATTACH_URETPROBE_CHECKED(skel, sym_name, prog_name) __ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name, true)
 
+/*
+ * -EFAULT in get_stackid normally means the stack-trace is not available,
+ * such as getting kernel stack trace in user mode
+ */
+#define STACK_ID_EFAULT(stack_id)	(stack_id == -EFAULT)
+
+#define STACK_ID_ERR(stack_id)		((stack_id < 0) && !STACK_ID_EFAULT(stack_id))
+
+/* hash collision (-EEXIST) suggests that stack map size may be too small */
+#define STACK_ID_COLLISION(stack_id)		(stack_id == -EEXIST)
+
 static void sig_handler(int signo);
 
 static long argp_parse_long(int key, const char *arg, struct argp_state *state);
@@ -205,6 +219,10 @@ static const struct argp_option argp_options[] = {
 	{"obj", 'O', "OBJECT", 0, "attach to allocator functions in the specified object", 0 },
 	{"percpu", 'P', NULL, 0, "trace percpu allocations", 0 },
 	{"symbols-prefix", 'S', "SYMBOLS_PREFIX", 0, "memory allocator symbols prefix", 0 },
+	{"stack-storage-size", OPT_STACK_STORAGE_SIZE, "STACK-STORAGE-SIZE", 0,
+	 "the number of unique stack traces that can be stored and displayed (default 10240)", 0 },
+	{"perf-max-stack-depth", OPT_PERF_MAX_STACK_DEPTH,
+	 "PERF-MAX-STACK-DEPTH", 0, "the limit for both kernel and user stack traces (default 127)", 0 },
 	{"verbose", 'v', NULL, 0, "verbose debug output", 0 },
 	{},
 };
@@ -274,7 +292,6 @@ int main(int argc, char *argv[])
 	printf("using page size: %ld\n", env.page_size);
 
 	env.kernel_trace = env.pid < 0 && !strlen(env.command);
-	printf("tracing kernel: %s\n", env.kernel_trace ? "true" : "false");
 
 	// if specific userspace program was specified,
 	// create the child process and use an eventfd to synchronize the call to exec()
@@ -358,6 +375,11 @@ int main(int argc, char *argv[])
 				env.perf_max_stack_depth * sizeof(unsigned long));
 	bpf_map__set_max_entries(skel->maps.stack_traces, env.stack_map_max_entries);
 
+	if (env.combined_only) {
+		bpf_map__set_max_entries(skel->maps.combined_allocs, COMBINED_ALLOCS_MAX_ENTRIES);
+		skel->rodata->combined_only = true;
+	}
+
 	// disable kernel tracepoints based on settings or availability
 	if (env.kernel_trace) {
 		if (!has_kernel_node_tracepoints())
@@ -382,12 +404,15 @@ int main(int argc, char *argv[])
 
 	// if userspace oriented, attach uprobes
 	if (!env.kernel_trace) {
+		printf("Attaching to pid %d, Ctrl+C to quit.\n", env.pid);
 		ret = attach_uprobes(skel);
 		if (ret) {
 			fprintf(stderr, "failed to attach uprobes\n");
 
 			goto cleanup;
 		}
+	} else {
+		printf("Attaching to kernel allocators, Ctrl+C to quit.\n");
 	}
 
 	ret = memleak_bpf__attach(skel);
@@ -439,8 +464,6 @@ int main(int argc, char *argv[])
 	}
 #endif
 
-	printf("Tracing outstanding memory allocs...  Hit Ctrl-C to end\n");
-
 	// main loop
 	while (!exiting && env.nr_intervals) {
 		env.nr_intervals--;
@@ -561,6 +584,22 @@ error_t argp_parse_arg(int key, char *arg, struct argp_state *state)
 	case 'v':
 		env.verbose = true;
 		break;
+	case OPT_PERF_MAX_STACK_DEPTH:
+		errno = 0;
+		env.perf_max_stack_depth = strtol(arg, NULL, 10);
+		if (errno) {
+			fprintf(stderr, "invalid perf max stack depth: %s\n", arg);
+			argp_usage(state);
+		}
+		break;
+	case OPT_STACK_STORAGE_SIZE:
+		errno = 0;
+		env.stack_map_max_entries = strtol(arg, NULL, 10);
+		if (errno) {
+			fprintf(stderr, "invalid stack storage size: %s\n", arg);
+			argp_usage(state);
+		}
+		break;
 	case ARGP_KEY_ARG:
 		pos_args++;
 
@@ -838,6 +877,8 @@ int print_outstanding_allocs(int allocs_fd, int stack_traces_fd)
 	struct tm *tm = localtime(&t);
 
 	size_t nr_allocs = 0;
+	size_t nr_missing_stacks = 0;
+	size_t nr_collision_stacks = 0;
 
 	// for each struct alloc_info "alloc_info" in the bpf map "allocs"
 	for (uint64_t prev_key = 0, curr_key = 0;; prev_key = curr_key) {
@@ -870,6 +911,10 @@ int print_outstanding_allocs(int allocs_fd, int stack_traces_fd)
 
 		// filter invalid stacks
 		if (alloc_info.stack_id < 0) {
+			/* handle stack id errors */
+			nr_missing_stacks += STACK_ID_ERR(alloc_info.stack_id);
+			nr_collision_stacks += STACK_ID_COLLISION(alloc_info.stack_id);
+
 			continue;
 		}
 
@@ -954,6 +999,13 @@ int print_outstanding_allocs(int allocs_fd, int stack_traces_fd)
 		}
 	}
 
+	if (nr_missing_stacks > 0) {
+		fprintf(stderr, "WARNING: %zu stack traces could not be displayed"
+			" due to memory shortage, including %zu caused by hash collisions."
+			" Consider increasing --stack-storage-size.\n",
+			nr_missing_stacks, nr_collision_stacks);
+	}
+
 	return 0;
 }
 
@@ -963,6 +1015,8 @@ int print_outstanding_combined_allocs(int combined_allocs_fd, int stack_traces_f
 	struct tm *tm = localtime(&t);
 
 	size_t nr_allocs = 0;
+	size_t nr_missing_stacks = 0;
+	size_t nr_collision_stacks = 0;
 
 	// for each stack_id "curr_key" and union combined_alloc_info "alloc"
 	// in bpf_map "combined_allocs"
@@ -996,6 +1050,16 @@ int print_outstanding_combined_allocs(int combined_allocs_fd, int stack_traces_f
 			.allocations = NULL
 		};
 
+		// filter invalid stacks
+		if (alloc.stack_id < 0) {
+			/* handle stack id errors */
+			if (STACK_ID_ERR(alloc.stack_id))
+				nr_missing_stacks += alloc.count;
+			if (STACK_ID_COLLISION(alloc.stack_id))
+				nr_collision_stacks += alloc.count;
+			continue;
+		}
+
 		memcpy(&allocs[nr_allocs], &alloc, sizeof(alloc));
 		nr_allocs++;
 	}
@@ -1010,6 +1074,13 @@ int print_outstanding_combined_allocs(int combined_allocs_fd, int stack_traces_f
 
 	print_stack_frames(allocs, nr_allocs, stack_traces_fd);
 
+	if (nr_missing_stacks > 0) {
+		fprintf(stderr, "WARNING: %zu stack traces could not be displayed"
+			" due to memory shortage, including %zu caused by hash collisions."
+			" Consider increasing --stack-storage-size.\n",
+			nr_missing_stacks, nr_collision_stacks);
+	}
+
 	return 0;
 }
 
